@@ -258,13 +258,13 @@ static void KafkaError_init (KafkaError *self,
                 self->str = NULL;
 }
 
-static int KafkaError_init0 (PyObject *selfobj, PyObject *args, 
+static int KafkaError_init0 (PyObject *selfobj, PyObject *args,
                              PyObject *kwargs) {
         KafkaError *self = (KafkaError *)selfobj;
         int code;
         int fatal = 0, retriable = 0, txn_requires_abort = 0;
         const char *reason = NULL;
-        static char *kws[] = { "error", "reason", "fatal", 
+        static char *kws[] = { "error", "reason", "fatal",
                                "retriable", "txn_requires_abort", NULL };
 
         if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|ziii", kws, &code,
@@ -310,7 +310,7 @@ static PyTypeObject KafkaErrorType = {
 	"  - Exceptions\n"
         "\n"
         "Args:\n"
-        "  error_code (KafkaError): Error code indicating the type of error.\n"
+        "  error (KafkaError): Error code indicating the type of error.\n"
         "\n"
         "  reason (str): Alternative message to describe the error.\n"
         "\n"
@@ -340,7 +340,6 @@ static PyTypeObject KafkaErrorType = {
         KafkaError_new             /* tp_new */
 };
 
-
 /**
  * @brief Internal factory to create KafkaError object.
  */
@@ -410,9 +409,6 @@ static void cfl_PyErr_Fatal (rd_kafka_resp_err_t err, const char *reason) {
 }
 
 
-
-
-
 /****************************************************************************
  *
  *
@@ -476,6 +472,13 @@ static PyObject *Message_offset (Message *self, PyObject *ignore) {
 		Py_RETURN_NONE;
 }
 
+static PyObject *Message_leader_epoch (Message *self, PyObject *ignore) {
+	if (self->leader_epoch >= 0)
+		return cfl_PyInt_FromInt(self->leader_epoch);
+	else
+		Py_RETURN_NONE;
+}
+
 
 static PyObject *Message_timestamp (Message *self, PyObject *ignore) {
 	return Py_BuildValue("iL",
@@ -571,6 +574,11 @@ static PyMethodDef Message_methods[] = {
 	  "  :rtype: int or None\n"
 	  "\n"
 	},
+        { "leader_epoch", (PyCFunction)Message_leader_epoch, METH_NOARGS,
+	  "  :returns: message offset leader epoch or None if not available.\n"
+	  "  :rtype: int or None\n"
+	  "\n"
+	},
 	{ "timestamp", (PyCFunction)Message_timestamp, METH_NOARGS,
           "Retrieve timestamp type and timestamp from message.\n"
           "The timestamp type is one of:\n\n"
@@ -743,7 +751,7 @@ PyTypeObject MessageType = {
 	0,		           /* tp_weaklistoffset */
 	0,		           /* tp_iter */
 	0,		           /* tp_iternext */
-	Message_methods,       /* tp_methods */
+	Message_methods,           /* tp_methods */
 	0,                         /* tp_members */
 	0,                         /* tp_getset */
 	0,                         /* tp_base */
@@ -784,6 +792,7 @@ PyObject *Message_new0 (const Handle *handle, const rd_kafka_message_t *rkm) {
 
 	self->partition = rkm->partition;
 	self->offset = rkm->offset;
+        self->leader_epoch = rd_kafka_message_leader_epoch(rkm);
 
 	self->timestamp = rd_kafka_message_timestamp(rkm, &self->tstype);
 
@@ -797,6 +806,158 @@ PyObject *Message_new0 (const Handle *handle, const rd_kafka_message_t *rkm) {
 
 
 
+/****************************************************************************
+ *
+ *
+ * Uuid
+ *
+ *
+ *
+ *
+ ****************************************************************************/
+static PyObject *Uuid_most_significant_bits (Uuid *self, PyObject *ignore) {
+        if(self->cUuid) {
+                return cfl_PyLong_FromLong(rd_kafka_Uuid_most_significant_bits(self->cUuid));
+        }
+        Py_RETURN_NONE;
+}
+
+static PyObject *Uuid_least_significant_bits (Uuid *self, PyObject *ignore) {
+        if(self->cUuid) {
+                return cfl_PyLong_FromLong(rd_kafka_Uuid_least_significant_bits(self->cUuid));
+        }
+        Py_RETURN_NONE;
+}
+
+static PyMethodDef Uuid_methods[] = {
+	{ "get_most_significant_bits", (PyCFunction)Uuid_most_significant_bits, METH_NOARGS,
+	  "  :returns: Most significant 64 bits of the 128 bits Uuid\n"
+	  "  :rtype: int\n"
+	  "\n"
+	},
+	{ "get_least_significant_bits", (PyCFunction)Uuid_least_significant_bits, METH_NOARGS,
+	  "  :returns: Least significant 64 bits of the 128 bits Uuid\n"
+	  "  :rtype: int\n"
+	  "\n"
+	},
+	{ NULL }
+};
+
+
+static PyObject *Uuid_str0 (Uuid *self) {
+        if(self->cUuid) {
+                const char *base64str = rd_kafka_Uuid_base64str(self->cUuid);
+                if(base64str)
+                        return cfl_PyUnistr(_FromString(base64str));
+        }
+        Py_RETURN_NONE;
+}
+
+static long Uuid_hash (Uuid *self) {
+        
+        return rd_kafka_Uuid_most_significant_bits(self->cUuid) ^ rd_kafka_Uuid_least_significant_bits(self->cUuid);
+}
+
+
+static PyObject *Uuid_new (PyTypeObject *type, PyObject *args,
+                               PyObject *kwargs) {
+        PyObject *self = type->tp_alloc(type, 1);
+        return self;
+}
+
+
+static int Uuid_init (PyObject *self0, PyObject *args,
+                          PyObject *kwargs) {
+        Uuid *self = (Uuid *)self0;
+        static char *kws[] = { "most_significant_bits",
+                               "least_significant_bits",
+                               NULL };
+        int64_t most_significant_bits;
+        int64_t least_significant_bits;
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs, "LL", kws,
+                                         &most_significant_bits, 
+                                         &least_significant_bits))
+                return -1;
+
+        self->cUuid = rd_kafka_Uuid_new(most_significant_bits, least_significant_bits);
+
+        return 0;
+}
+
+static int Uuid_clear (Uuid *self) {
+	if (self->cUuid) {
+		rd_kafka_Uuid_destroy(self->cUuid);
+		self->cUuid = NULL;
+	}
+	return 0;
+}
+
+static int Uuid_traverse (Uuid *self,
+                        visitproc visit, void *arg) {
+        return 0;
+}
+
+static void Uuid_dealloc (Uuid *self) {
+        PyObject_GC_UnTrack(self);
+        Uuid_clear(self);
+
+        Py_TYPE(self)->tp_free((PyObject *)self);
+}
+
+PyTypeObject UuidType = {
+        PyVarObject_HEAD_INIT(NULL, 0)
+        "cimpl.Uuid",                           /* tp_name */
+        sizeof(Uuid),                           /* tp_basicsize */
+        0,                                      /* tp_itemsize */
+        (destructor)Uuid_dealloc,               /* tp_dealloc */
+        0,                                      /* tp_print */
+        0,                                      /* tp_getattr */
+        0,                                      /* tp_setattr */
+        0,                                      /* tp_compare */
+        (reprfunc)Uuid_str0,                    /* tp_repr */
+        0,                                      /* tp_as_number */
+        0,                                      /* tp_as_sequence */
+        0,                                      /* tp_as_mapping */
+        (hashfunc)Uuid_hash,                    /* tp_hash */
+        0,                                      /* tp_call */
+        0,                                      /* tp_str */
+        PyObject_GenericGetAttr,                /* tp_getattro */
+        0,                                      /* tp_setattro */
+        0,                                      /* tp_as_buffer */
+        Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | 
+	Py_TPFLAGS_HAVE_GC,                     /* tp_flags */
+        "Generic Uuid. Being used in various identifiers including topic_id.\n"
+        "\n"
+        ".. py:function:: Uuid(most_significant_bits, least_significant_bits)\n"
+        "\n"
+        "  Instantiate a Uuid object.\n"
+        "\n"
+        "  :param long most_significant_bits: Most significant 64 bits of the 128 bits Uuid.\n"
+        "  :param long least_significant_bits: Least significant 64 bits of the 128 bits Uuid.\n"
+        "  :rtype: Uuid\n"
+        "\n"
+"\n",                                           /* tp_doc */
+        (traverseproc)Uuid_traverse,            /* tp_traverse */
+        (inquiry)Uuid_clear,                    /* tp_clear */
+        0,                                      /* tp_richcompare */
+        0,                                      /* tp_weaklistoffset */
+        0,                                      /* tp_iter */
+        0,                                      /* tp_iternext */
+        Uuid_methods,                           /* tp_methods */
+        0,                                      /* tp_members */
+        0,                                      /* tp_getset */
+        0,                                      /* tp_base */
+        0,                                      /* tp_dict */
+        0,                                      /* tp_descr_get */
+        0,                                      /* tp_descr_set */
+        0,                                      /* tp_dictoffset */
+        Uuid_init,                              /* tp_init */
+        0,                                      /* tp_alloc */
+        Uuid_new                                /* tp_new */
+};
+
+
 
 /****************************************************************************
  *
@@ -816,15 +977,32 @@ static int TopicPartition_clear (TopicPartition *self) {
 		Py_DECREF(self->error);
 		self->error = NULL;
 	}
+	if (self->metadata) {
+		free(self->metadata);
+		self->metadata = NULL;
+	}
 	return 0;
 }
 
 static void TopicPartition_setup (TopicPartition *self, const char *topic,
 				  int partition, long long offset,
+                                  int32_t leader_epoch,
+				  const char *metadata,
 				  rd_kafka_resp_err_t err) {
 	self->topic = strdup(topic);
 	self->partition = partition;
 	self->offset = offset;
+
+        if (leader_epoch < 0)
+                leader_epoch = -1;
+        self->leader_epoch = leader_epoch;
+
+	if (metadata != NULL) {
+		self->metadata = strdup(metadata);
+	} else {
+		self->metadata = NULL;
+	}
+
 	self->error = KafkaError_new_or_None(err, NULL);
 }
 
@@ -842,19 +1020,27 @@ static int TopicPartition_init (PyObject *self, PyObject *args,
 				      PyObject *kwargs) {
 	const char *topic;
 	int partition = RD_KAFKA_PARTITION_UA;
+        int32_t leader_epoch = -1;
 	long long offset = RD_KAFKA_OFFSET_INVALID;
+	const char *metadata = NULL;
+
 	static char *kws[] = { "topic",
 			       "partition",
 			       "offset",
+			       "metadata",
+                               "leader_epoch",
 			       NULL };
 
-	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|iL", kws,
-					 &topic, &partition, &offset))
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|iLsi", kws,
+					 &topic, &partition, &offset,
+					 &metadata,
+                                         &leader_epoch)) {
 		return -1;
+	}
 
 	TopicPartition_setup((TopicPartition *)self,
-			     topic, partition, offset, 0);
-
+			     topic, partition, offset,
+                             leader_epoch, metadata, 0);
 	return 0;
 }
 
@@ -874,6 +1060,13 @@ static int TopicPartition_traverse (TopicPartition *self,
 	return 0;
 }
 
+static PyObject *TopicPartition_get_leader_epoch (TopicPartition *tp, void *closure) {
+        if (tp->leader_epoch >= 0) {
+                return cfl_PyInt_FromInt(tp->leader_epoch);
+        }
+        Py_RETURN_NONE;
+}
+
 
 static PyMemberDef TopicPartition_members[] = {
         { "topic", T_STRING, offsetof(TopicPartition, topic), READONLY,
@@ -889,11 +1082,28 @@ static PyMemberDef TopicPartition_members[] = {
           " :py:const:`OFFSET_STORED`,"
           " :py:const:`OFFSET_INVALID`\n"
         },
+        {"metadata", T_STRING, offsetof(TopicPartition, metadata), READONLY,
+         "attribute metadata: Optional application metadata committed with the "
+         "offset (string)"},
         { "error", T_OBJECT, offsetof(TopicPartition, error), READONLY,
           ":attribute error: Indicates an error (with :py:class:`KafkaError`) unless None." },
         { NULL }
 };
 
+static PyGetSetDef TopicPartition_getters_and_setters[] = {
+        {
+          /* name */
+          "leader_epoch",
+          (getter) TopicPartition_get_leader_epoch,
+          NULL,
+          /* doc */
+          ":attribute leader_epoch: Offset leader epoch (int), or None",
+          /* closure */
+          NULL
+        },
+        { NULL }
+};
+
 
 static PyObject *TopicPartition_str0 (TopicPartition *self) {
         PyObject *errstr = NULL;
@@ -901,8 +1111,15 @@ static PyObject *TopicPartition_str0 (TopicPartition *self) {
         const char *c_errstr = NULL;
 	PyObject *ret;
 	char offset_str[40];
+        char leader_epoch_str[12];
 
 	snprintf(offset_str, sizeof(offset_str), "%"CFL_PRId64"", self->offset);
+        if (self->leader_epoch >= 0)
+                snprintf(leader_epoch_str, sizeof(leader_epoch_str),
+                        "%"CFL_PRId32"", self->leader_epoch);
+        else
+                snprintf(leader_epoch_str, sizeof(leader_epoch_str),
+                        "None");
 
         if (self->error != Py_None) {
                 errstr = cfl_PyObject_Unistr(self->error);
@@ -911,9 +1128,10 @@ static PyObject *TopicPartition_str0 (TopicPartition *self) {
 
 	ret = cfl_PyUnistr(
 		_FromFormat("TopicPartition{topic=%s,partition=%"CFL_PRId32
-			    ",offset=%s,error=%s}",
+			    ",offset=%s,leader_epoch=%s,error=%s}",
 			    self->topic, self->partition,
 			    offset_str,
+                            leader_epoch_str,
 			    c_errstr ? c_errstr : "None"));
         Py_XDECREF(errstr8);
         Py_XDECREF(errstr);
@@ -1005,53 +1223,66 @@ PyTypeObject TopicPartitionType = {
 	"It is typically used to provide a list of topics or partitions for "
 	"various operations, such as :py:func:`Consumer.assign()`.\n"
 	"\n"
-	".. py:function:: TopicPartition(topic, [partition], [offset])\n"
+	".. py:function:: TopicPartition(topic, [partition], [offset],"
+        " [metadata], [leader_epoch])\n"
 	"\n"
 	"  Instantiate a TopicPartition object.\n"
 	"\n"
 	"  :param string topic: Topic name\n"
 	"  :param int partition: Partition id\n"
 	"  :param int offset: Initial partition offset\n"
+        "  :param string metadata: Offset metadata\n"
+        "  :param int leader_epoch: Offset leader epoch\n"
 	"  :rtype: TopicPartition\n"
 	"\n"
 	"\n", /*tp_doc*/
 	(traverseproc)TopicPartition_traverse, /* tp_traverse */
 	(inquiry)TopicPartition_clear,       /* tp_clear */
 	(richcmpfunc)TopicPartition_richcompare, /* tp_richcompare */
-	0,		           /* tp_weaklistoffset */
-	0,		           /* tp_iter */
-	0,		           /* tp_iternext */
-	0,                         /* tp_methods */
-	TopicPartition_members,/* tp_members */
-	0,                         /* tp_getset */
-	0,                         /* tp_base */
-	0,                         /* tp_dict */
-	0,                         /* tp_descr_get */
-	0,                         /* tp_descr_set */
-	0,                         /* tp_dictoffset */
-	TopicPartition_init,       /* tp_init */
-	0,                         /* tp_alloc */
-	TopicPartition_new         /* tp_new */
+	0,		                    /* tp_weaklistoffset */
+	0,		                    /* tp_iter */
+	0,		                    /* tp_iternext */
+	0,                                  /* tp_methods */
+	TopicPartition_members,             /* tp_members */
+	TopicPartition_getters_and_setters, /* tp_getset */
+	0,                                  /* tp_base */
+	0,                                  /* tp_dict */
+	0,                                  /* tp_descr_get */
+	0,                                  /* tp_descr_set */
+	0,                                  /* tp_dictoffset */
+	TopicPartition_init,                /* tp_init */
+	0,                                  /* tp_alloc */
+	TopicPartition_new                  /* tp_new */
 };
 
 /**
  * @brief Internal factory to create a TopicPartition object.
  */
 static PyObject *TopicPartition_new0 (const char *topic, int partition,
-				      long long offset,
+				      long long offset, int32_t leader_epoch,
+                                      const char *metadata,
 				      rd_kafka_resp_err_t err) {
 	TopicPartition *self;
 
 	self = (TopicPartition *)TopicPartitionType.tp_new(
 		&TopicPartitionType, NULL, NULL);
 
-	TopicPartition_setup(self, topic, partition, offset, err);
+	TopicPartition_setup(self, topic, partition,
+			     offset, leader_epoch,
+                             metadata, err);
 
 	return (PyObject *)self;
 }
 
 
 
+PyObject *c_part_to_py(const rd_kafka_topic_partition_t *rktpar) {
+        return TopicPartition_new0(rktpar->topic, rktpar->partition,
+			           rktpar->offset,
+                                   rd_kafka_topic_partition_get_leader_epoch(rktpar),
+				   rktpar->metadata,
+				   rktpar->err);
+}
 
 /**
  * @brief Convert C rd_kafka_topic_partition_list_t to Python list(TopicPartition).
@@ -1066,10 +1297,7 @@ PyObject *c_parts_to_py (const rd_kafka_topic_partition_list_t *c_parts) {
 
 	for (i = 0 ; i < (size_t)c_parts->cnt ; i++) {
 		const rd_kafka_topic_partition_t *rktpar = &c_parts->elems[i];
-		PyList_SET_ITEM(parts, i,
-				TopicPartition_new0(
-					rktpar->topic, rktpar->partition,
-					rktpar->offset, rktpar->err));
+		PyList_SET_ITEM(parts, i, c_part_to_py(rktpar));
 	}
 
 	return parts;
@@ -1094,6 +1322,7 @@ rd_kafka_topic_partition_list_t *py_to_c_parts (PyObject *plist) {
 	c_parts = rd_kafka_topic_partition_list_new((int)PyList_Size(plist));
 
 	for (i = 0 ; i < (size_t)PyList_Size(plist) ; i++) {
+		rd_kafka_topic_partition_t *rktpar;
 		TopicPartition *tp = (TopicPartition *)
 			PyList_GetItem(plist, i);
 
@@ -1106,15 +1335,60 @@ rd_kafka_topic_partition_list_t *py_to_c_parts (PyObject *plist) {
 			return NULL;
 		}
 
-		rd_kafka_topic_partition_list_add(c_parts,
-						  tp->topic,
-						  tp->partition)->offset =
-			tp->offset;
+		rktpar = rd_kafka_topic_partition_list_add(c_parts,
+							   tp->topic,
+							   tp->partition);
+		rktpar->offset = tp->offset;
+                rd_kafka_topic_partition_set_leader_epoch(rktpar,
+                        tp->leader_epoch);
+		if (tp->metadata != NULL) {
+			rktpar->metadata_size = strlen(tp->metadata) + 1;
+			rktpar->metadata = strdup(tp->metadata);
+		} else {
+			rktpar->metadata_size = 0;
+			rktpar->metadata = NULL;
+		}
 	}
 
 	return c_parts;
 }
 
+/**
+ * @brief Convert C rd_kafka_topic_partition_result_t to Python dict(TopicPartition, KafkaException).
+ * 
+ * @returns The new Python dict object.
+ */
+PyObject *c_topic_partition_result_to_py_dict(
+    const rd_kafka_topic_partition_result_t **partition_results,
+    size_t cnt) {
+        PyObject *result = NULL;
+        size_t i;
+
+        result = PyDict_New();
+
+        for (i = 0; i < cnt; i++) {
+                PyObject *key;
+                PyObject *value;
+                const rd_kafka_topic_partition_t *c_topic_partition;
+                const rd_kafka_error_t *c_error;
+
+                c_topic_partition =
+                    rd_kafka_topic_partition_result_partition(partition_results[i]);
+                c_error = rd_kafka_topic_partition_result_error(partition_results[i]);
+
+                value = KafkaError_new_or_None(rd_kafka_error_code(c_error),
+                                               rd_kafka_error_string(c_error));
+                key = c_part_to_py(c_topic_partition);
+               
+                PyDict_SetItem(result, key, value);
+
+                Py_DECREF(key);
+                Py_DECREF(value);
+        }
+
+        return result;
+}
+
 #ifdef RD_KAFKA_V_HEADERS
 
 
@@ -1288,7 +1562,7 @@ PyObject *c_headers_to_py (rd_kafka_headers_t *headers) {
 
     while (!rd_kafka_header_get_all(headers, idx++,
                                      &header_key, &header_value, &header_value_size)) {
-            // Create one (key, value) tuple for each header
+            /* Create one (key, value) tuple for each header */
             PyObject *header_tuple = PyTuple_New(2);
             PyTuple_SetItem(header_tuple, 0,
                 cfl_PyUnistr(_FromString(header_key))
@@ -1299,6 +1573,7 @@ PyObject *c_headers_to_py (rd_kafka_headers_t *headers) {
                         cfl_PyBin(_FromStringAndSize(header_value, header_value_size))
                     );
             } else {
+                Py_INCREF(Py_None);
                 PyTuple_SetItem(header_tuple, 1, Py_None);
             }
         PyList_SET_ITEM(header_list, idx-1, header_tuple);
@@ -1357,6 +1632,91 @@ rd_kafka_consumer_group_metadata_t *py_to_c_cgmd (PyObject *obj) {
         return cgmd;
 }
 
+PyObject *c_Node_to_py(const rd_kafka_Node_t *c_node) {
+        PyObject *node = NULL;
+        PyObject *Node_type = NULL;
+        PyObject *args = NULL;
+        PyObject *kwargs = NULL;
+        const char *rack = NULL;
+
+        if(!c_node)
+                Py_RETURN_NONE;
+
+        Node_type = cfl_PyObject_lookup("confluent_kafka",
+                                        "Node");
+        if (!Node_type) {
+                goto err;
+        }
+
+        kwargs = PyDict_New();
+
+        cfl_PyDict_SetInt(kwargs, "id", rd_kafka_Node_id(c_node));
+        cfl_PyDict_SetInt(kwargs, "port", rd_kafka_Node_port(c_node));
+        if (rd_kafka_Node_host(c_node))
+                cfl_PyDict_SetString(kwargs, "host", rd_kafka_Node_host(c_node));
+        else
+                PyDict_SetItemString(kwargs, "host", Py_None);
+        if((rack = rd_kafka_Node_rack(c_node)))
+                cfl_PyDict_SetString(kwargs, "rack", rack);
+
+        args = PyTuple_New(0);
+
+        node = PyObject_Call(Node_type, args, kwargs);
+
+        Py_DECREF(Node_type);
+        Py_DECREF(args);
+        Py_DECREF(kwargs);
+        return node;
+
+err:
+        Py_XDECREF(Node_type);
+        Py_XDECREF(args);
+        Py_XDECREF(kwargs);
+        Py_XDECREF(node);
+        return NULL;
+}
+
+PyObject *c_Uuid_to_py(const rd_kafka_Uuid_t *c_uuid) {
+        PyObject *uuid = NULL;
+        PyObject *Uuid_type = NULL;
+        PyObject *args = NULL;
+        PyObject *kwargs = NULL;
+        
+        if(!c_uuid)
+                Py_RETURN_NONE;
+
+        Uuid_type = cfl_PyObject_lookup("confluent_kafka",
+                                        "Uuid");
+        if (!Uuid_type) {
+                goto err;
+        }
+
+        
+        kwargs = PyDict_New();
+
+        cfl_PyDict_SetLong(kwargs, "most_significant_bits", rd_kafka_Uuid_most_significant_bits(c_uuid));
+        cfl_PyDict_SetLong(kwargs, "least_significant_bits", rd_kafka_Uuid_least_significant_bits(c_uuid));
+
+        args = PyTuple_New(0);
+
+        uuid = PyObject_Call(Uuid_type, args, kwargs);
+
+        Py_DECREF(Uuid_type);
+        Py_DECREF(args);
+        Py_DECREF(kwargs);
+
+        return uuid;
+
+err:
+        Py_XDECREF(Uuid_type);
+        Py_XDECREF(args);
+        Py_XDECREF(kwargs);
+        Py_XDECREF(uuid);
+
+        return NULL;
+
+}
+
 
 /****************************************************************************
  *
@@ -1521,6 +1881,62 @@ static void log_cb (const rd_kafka_t *rk, int level,
         CallState_resume(cs);
 }
 
+/**
+ * @brief Translate Python \p key and \p value to C types and set on
+ *        provided \p extensions char* array at the provided index.
+ *
+ * @returns 1 on success or 0 if an exception was raised.
+ */
+static int py_extensions_to_c (char **extensions, Py_ssize_t idx,
+                               PyObject *key, PyObject *value) {
+        PyObject *ks, *ks8, *vo8 = NULL;
+        const char *k;
+        const char *v;
+        Py_ssize_t ksize = 0;
+        Py_ssize_t vsize = 0;
+
+        if (!(ks = cfl_PyObject_Unistr(key))) {
+                PyErr_SetString(PyExc_TypeError,
+                                "expected extension key to be unicode "
+                                "string");
+                return 0;
+        }
+
+        k = cfl_PyUnistr_AsUTF8(ks, &ks8);
+        ksize = (Py_ssize_t)strlen(k);
+
+        if (cfl_PyUnistr(_Check(value))) {
+                /* Unicode string, translate to utf-8. */
+                v = cfl_PyUnistr_AsUTF8(value, &vo8);
+                if (!v) {
+                        Py_DECREF(ks);
+                        Py_XDECREF(ks8);
+                        return 0;
+                }
+                vsize = (Py_ssize_t)strlen(v);
+        } else {
+                PyErr_Format(PyExc_TypeError,
+                             "expected extension value to be "
+                             "unicode string, not %s",
+                             ((PyTypeObject *)PyObject_Type(value))->
+                             tp_name);
+                Py_DECREF(ks);
+                Py_XDECREF(ks8);
+                return 0;
+        }
+
+        extensions[idx] = (char*)malloc(ksize + 1);
+        snprintf(extensions[idx], ksize + 1, "%s", k);
+        extensions[idx + 1] = (char*)malloc(vsize + 1);
+        snprintf(extensions[idx + 1], vsize + 1, "%s", v);
+
+        Py_DECREF(ks);
+        Py_XDECREF(ks8);
+        Py_XDECREF(vo8);
+
+        return 1;
+}
+
 static void oauth_cb (rd_kafka_t *rk, const char *oauthbearer_config,
                       void *opaque) {
         Handle *h = opaque;
@@ -1528,6 +1944,10 @@ static void oauth_cb (rd_kafka_t *rk, const char *oauthbearer_config,
         CallState *cs;
         const char *token;
         double expiry;
+        const char *principal = "";
+        PyObject *extensions = NULL;
+        char **rd_extensions = NULL;
+        Py_ssize_t rd_extensions_size = 0;
         char err_msg[2048];
         rd_kafka_resp_err_t err_code;
 
@@ -1538,26 +1958,58 @@ static void oauth_cb (rd_kafka_t *rk, const char *oauthbearer_config,
         Py_DECREF(eo);
 
         if (!result) {
-                goto err;
+                goto fail;
         }
-        if (!PyArg_ParseTuple(result, "sd", &token, &expiry)) {
+        if (!PyArg_ParseTuple(result, "sd|sO!", &token, &expiry, &principal, &PyDict_Type, &extensions)) {
                 Py_DECREF(result);
-                PyErr_Format(PyExc_TypeError,
+                PyErr_SetString(PyExc_TypeError,
                              "expect returned value from oauth_cb "
-                             "to be (token_str, expiry_time) tuple");
+                             "to be (token_str, expiry_time[, principal, extensions]) tuple");
                 goto err;
         }
+
+        if (extensions) {
+                int len = (int)PyDict_Size(extensions);
+                rd_extensions = (char **)malloc(2 * len * sizeof(char *));
+                Py_ssize_t pos = 0;
+                PyObject *ko, *vo;
+                while (PyDict_Next(extensions, &pos, &ko, &vo)) {
+                        if (!py_extensions_to_c(rd_extensions, rd_extensions_size, ko, vo)) {
+                                Py_DECREF(result);
+                                free(rd_extensions);
+                                goto err;
+                        }
+                        rd_extensions_size = rd_extensions_size + 2;
+                }
+        }
+
         err_code = rd_kafka_oauthbearer_set_token(h->rk, token,
                                                   (int64_t)(expiry * 1000),
-                                                  "", NULL, 0, err_msg,
+                                                  principal, (const char **)rd_extensions, rd_extensions_size, err_msg,
                                                   sizeof(err_msg));
         Py_DECREF(result);
-        if (err_code) {
+        if (rd_extensions) {
+                int i;
+                for(i = 0; i < rd_extensions_size; i++) {
+                        free(rd_extensions[i]);
+                }
+                free(rd_extensions);
+        }
+
+        if (err_code != RD_KAFKA_RESP_ERR_NO_ERROR) {
                 PyErr_Format(PyExc_ValueError, "%s", err_msg);
-                goto err;
+                goto fail;
         }
         goto done;
 
+fail:
+        err_code = rd_kafka_oauthbearer_set_token_failure(h->rk, "OAuth callback raised exception");
+        if (err_code != RD_KAFKA_RESP_ERR_NO_ERROR) {
+                PyErr_SetString(PyExc_ValueError, "Failed to set token failure");
+                goto err;
+        }
+        PyErr_Clear();
+        goto done;
  err:
         CallState_crash(cs);
         rd_kafka_yield(h->rk);
@@ -1848,7 +2300,7 @@ rd_kafka_conf_t *common_conf_setup (rd_kafka_type_t ktype,
                 PyDict_Update(confdict, kwargs);
         }
 
-        if (ktype == RD_KAFKA_CONSUMER && 
+        if (ktype == RD_KAFKA_CONSUMER &&
                 !PyDict_GetItemString(confdict, "group.id")) {
 
                 PyErr_SetString(PyExc_ValueError,
@@ -2225,6 +2677,11 @@ void cfl_PyDict_SetInt (PyObject *dict, const char *name, int val) {
         Py_DECREF(vo);
 }
 
+void cfl_PyDict_SetLong (PyObject *dict, const char *name, long val) {
+        PyObject *vo = cfl_PyLong_FromLong(val);
+        PyDict_SetItemString(dict, name, vo);
+        Py_DECREF(vo);
+}
 
 int cfl_PyObject_SetString (PyObject *o, const char *name, const char *val) {
         PyObject *vo = cfl_PyUnistr(_FromString(val));
@@ -2254,7 +2711,7 @@ int cfl_PyObject_SetInt (PyObject *o, const char *name, int val) {
  */
 int cfl_PyObject_GetAttr (PyObject *object, const char *attr_name,
                           PyObject **valp, const PyTypeObject *py_type,
-                          int required) {
+                          int required, int allow_None) {
         PyObject *o;
 
         o = PyObject_GetAttrString(object, attr_name);
@@ -2269,7 +2726,7 @@ int cfl_PyObject_GetAttr (PyObject *object, const char *attr_name,
                 return 0;
         }
 
-        if (py_type && Py_TYPE(o) != py_type) {
+        if (!(allow_None && o == Py_None) && py_type && Py_TYPE(o) != py_type) {
                 Py_DECREF(o);
                 PyErr_Format(PyExc_TypeError,
                              "Expected .%s to be %s type, not %s",
@@ -2300,7 +2757,7 @@ int cfl_PyObject_GetInt (PyObject *object, const char *attr_name, int *valp,
 #else
                                   &PyInt_Type,
 #endif
-                                  required))
+                                  required, 0))
                 return 0;
 
         if (!o) {
@@ -2336,17 +2793,17 @@ int cfl_PyBool_get (PyObject *object, const char *name, int *valp) {
         return 1;
 }
 
-
 /**
  * @brief Get attribute \p attr_name from \p object and make sure it is
- *        a string type.
+ *        a string type or None if \p allow_None is 1
  *
  * @returns 1 if \p valp was updated with a newly allocated copy of either the
- *          object value (UTF8), or \p defval.
+ *          object value (UTF8), or \p defval or NULL if the attr is None
  *          0 if an exception was raised.
  */
 int cfl_PyObject_GetString (PyObject *object, const char *attr_name,
-                            char **valp, const char *defval, int required) {
+                            char **valp, const char *defval, int required,
+                            int allow_None) {
         PyObject *o, *uo, *uop;
 
         if (!cfl_PyObject_GetAttr(object, attr_name, &o,
@@ -2358,7 +2815,7 @@ int cfl_PyObject_GetString (PyObject *object, const char *attr_name,
                                    *           proper conversion below. */
                                   NULL,
 #endif
-                                  required))
+                                  required, allow_None))
                 return 0;
 
         if (!o) {
@@ -2366,6 +2823,12 @@ int cfl_PyObject_GetString (PyObject *object, const char *attr_name,
                 return 1;
         }
 
+        if (o == Py_None) {
+                Py_DECREF(o);
+                *valp = NULL;
+                return 1;
+        }
+
         if (!(uo = cfl_PyObject_Unistr(o))) {
                 Py_DECREF(o);
                 PyErr_Format(PyExc_TypeError,
@@ -2411,6 +2874,55 @@ PyObject *cfl_int32_array_to_py_list (const int32_t *arr, size_t cnt) {
 }
 
 
+/****************************************************************************
+ *
+ *
+ * Methods common across all types of clients.
+ *
+ *
+ *
+ *
+ ****************************************************************************/
+
+const char set_sasl_credentials_doc[] = PyDoc_STR(
+        ".. py:function:: set_sasl_credentials(username, password)\n"
+        "\n"
+        "  Sets the SASL credentials used for this client.\n"
+        "  These credentials will overwrite the old ones, and will be used the next time the client needs to authenticate.\n"
+        "  This method will not disconnect existing broker connections that have been established with the old credentials.\n"
+        "  This method is applicable only to SASL PLAIN and SCRAM mechanisms.\n");
+
+
+PyObject *set_sasl_credentials(Handle *self, PyObject *args, PyObject *kwargs) {
+        const char *username = NULL;
+        const char *password = NULL;
+        rd_kafka_error_t* error;
+        CallState cs;
+        static char *kws[] = {"username", "password", NULL};
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ss", kws,
+                                         &username, &password)) {
+                return NULL;
+        }
+
+        CallState_begin(self, &cs);
+        error = rd_kafka_sasl_set_credentials(self->rk, username, password);
+
+        if (!CallState_end(self, &cs)) {
+                if (error) /* Ignore error in favour of callstate exception */
+                        rd_kafka_error_destroy(error);
+                return NULL;
+        }
+
+        if (error) {
+                cfl_PyErr_from_error_destroy(error);
+                return NULL;
+        }
+
+        Py_RETURN_NONE;
+}
+
+
 /****************************************************************************
  *
  *
@@ -2523,7 +3035,7 @@ static char *KafkaError_add_errs (PyObject *dict, const char *origdoc) {
 
 	_PRINT("\n");
 
-	return doc; // FIXME: leak
+	return doc; /* FIXME: leak */
 }
 
 
@@ -2541,12 +3053,18 @@ static struct PyModuleDef cimpl_moduledef = {
 static PyObject *_init_cimpl (void) {
 	PyObject *m;
 
+/* PyEval_InitThreads became deprecated in Python 3.9 and will be removed in Python 3.11.
+ * Prior to Python 3.7, this call was required to initialize the GIL. */
+#if PY_VERSION_HEX < 0x03090000
         PyEval_InitThreads();
+#endif
 
 	if (PyType_Ready(&KafkaErrorType) < 0)
 		return NULL;
 	if (PyType_Ready(&MessageType) < 0)
 		return NULL;
+	if (PyType_Ready(&UuidType) < 0)
+		return NULL;
 	if (PyType_Ready(&TopicPartitionType) < 0)
 		return NULL;
 	if (PyType_Ready(&ProducerType) < 0)
@@ -2576,6 +3094,9 @@ static PyObject *_init_cimpl (void) {
 	Py_INCREF(&MessageType);
 	PyModule_AddObject(m, "Message", (PyObject *)&MessageType);
 
+        Py_INCREF(&UuidType);
+        PyModule_AddObject(m, "Uuid", (PyObject *)&UuidType);
+
 	Py_INCREF(&TopicPartitionType);
 	PyModule_AddObject(m, "TopicPartition",
 			   (PyObject *)&TopicPartitionType);
